from warnings import warn
from functools import wraps

from metagen.config.config import load_yaml, BASE_CONFIG_FILE, Config
from metagen.register import register_factory, RegisterABC
from metagen.generator import JSONSerializer, JSONDeserializer, _Generator
from metagen.importer import Importer

CONFIG = Config(**load_yaml(BASE_CONFIG_FILE))

# register
register = register_factory.get(CONFIG.register_setting.registerName)()


# helpers dacoratoes
def exist_in_register(element):
    @wraps(element)
    def checking_register(*args, **kwargs):
        instance = element(*args, **kwargs)
        if register.check_register(instance):
            registered_element = register.get_by_hash(hash(instance))
            warn(str(f'Element duplication: Element {instance.__class__.__name__} with parameters: '
                 f'{"; ".join([f"{k}: {v}" for k, v in kwargs.items()])} found in register. Element '
                 f'{registered_element.__repr__()} returned instead'))
            return registered_element
        else:
            register.add(instance)
            return instance
    return checking_register


serializer = JSONSerializer(reg=register)

from metagen.metadata import element_factory
deserializer = JSONDeserializer(factory=element_factory)
importer = Importer(**CONFIG.importer_setting.dict())


def Generator(register=register) -> _Generator:
    return _Generator(serializer=serializer, deserializer=deserializer, importer=importer, reg=register)




from metagen.helpers import Pipe, is_exist, str2Path


class PathCheck(Pipe):
    """helper class provided converting str input to Path and check if path exist"""
    str2Path = str2Path
    is_exist = is_exist

path_check = PathCheck()


import pytest
from metagen import Generator
from metagen import LayerTemplate


def test_gen_get_types():
    gen = Generator()
    lt1 = LayerTemplate(applicationKey='app', nameInternal=f'lt', nameDisplay='lt')
    lt2 = LayerTemplate(applicationKey='app', nameInternal=f'lt', nameDisplay='lt')
    lts = gen.get_elements_by_type(LayerTemplate)


from pydantic import BaseModel, Field, validator
from typing import List, Union, Optional, Type
from uuid import UUID
from datetime import datetime

from metagen.base import LeafABC, set_key_from_input
from metagen.components._general import Component
from metagen.components._general import Filter



# TODO: Verify layerState - application True
# TODO: how exactly works and by generated LayerState
# TODO: Timeline legent generalize
# TODO: Hashble TimilineItem, TimelineLayers check for duplicity
# TODO: add style definition to raster


class TimelinePeriod(BaseModel):
    start: Union[str, datetime]
    end: Union[str, datetime]


class LayerState(BaseModel):
    layerTemplateKey: Union[str, UUID, Type[LeafABC]]
    styleKey: Optional[Union[str, UUID, Type[LeafABC]]]
    filterByActive: dict = Field(default={"application": True})

    _set_key = validator('layerTemplateKey', pre=True, allow_reuse=True)(set_key_from_input)


class TimelineLegend(BaseModel):
    layerTemplateKey: Union[str, UUID, Type[LeafABC]]

    _set_key = validator('layerTemplateKey', pre=True, allow_reuse=True)(set_key_from_input)


class TimelineItemPeriod(BaseModel):
    filter: Filter
    filterByActive: dict = Field(default={"application": True})


class TimelineItem(BaseModel):
    mapZIndex: int
    periods: TimelineItemPeriod
    layerState: LayerState

    @classmethod
    def set_by_layerTemplate(cls, mapZIndex: int, layerTemplate: Type[LeafABC], style: Optional[Type[LeafABC]] = None):
        filter = Filter.set('layerTemplateKey', layerTemplate)
        periods = TimelineItemPeriod(filter=filter)
        layerstate = LayerState(layerTemplateKey=layerTemplate, styleKey=style)
        return cls(mapZIndex=mapZIndex, periods=periods, layerState=layerstate)


class TimelineLayer(BaseModel):
    legend: Optional[TimelineLegend]
    items: List[TimelineItem] = Field(default=[])

    @classmethod
    def set(cls, legend: TimelineLegend, item: Union[TimelineItem, List[TimelineItem]]):
        if isinstance(item, TimelineItem):
            item = [item]
        return cls(legend=legend, items=item)

    def set_legend(self, legend: TimelineLegend):
        self.legend = legend

    def add_item(self, item: Union[TimelineItem, List[TimelineItem]]):
        if isinstance(item, TimelineItem):
            self.items.append(item)
        elif isinstance(item, list):
            self.items += item


class Timeline(Component):
    timelinePeriod: Optional[TimelinePeriod]
    timelineLayers: List[TimelineLayer] = Field(default=[])

    @classmethod
    def set(cls, start: Union['str', datetime],
            end: Union['str', datetime],
            timelineLayers: Union[TimelineLayer, List[TimelineLayer]] = []):

        if isinstance(timelineLayers, TimelineLayer):
            timelineLayers = [timelineLayers]

        return cls(timelinePeriod=TimelinePeriod(start=start, end=end), timelineLayers=timelineLayers)

    def set_timePeriod(self, start: Union['str', datetime], end: Union['str', datetime]) -> None:
        """Set timeline time period range"""
        self.timelinePeriod = TimelinePeriod(start=start, end=end)

    def add_timelineLayer(self, timelineLayer: Union[TimelineLayer, List[TimelineLayer]]) -> None:
        """add TimeLayer into timelayer list"""
        if isinstance(timelineLayer, TimelineLayer):
            self.timelineLayers.append(timelineLayer)
        elif isinstance(timelineLayer, list):
            self.timelineLayers += timelineLayer

    def set_timelineLayer_by_layerTemplate(self, legend: Type[LeafABC], items: List[Type[LeafABC]] ):
        """
        Work around to set and ad timelineLayer by layerTempalte. LayerTemplates as items are add into the timiline
        layer in order of the layertemplates listd. mapZindex is made in same order
        """
        timeline_layer = TimelineLayer(legend=TimelineLegend(layerTemplateKey=legend))
        for index, layer_template in enumerate(items):
            timeline_layer.add_item(TimelineItem.set_by_layerTemplate(mapZIndex=index, layerTemplate=layer_template))
        self.timelineLayers.append(timeline_layer)


import pytest

from metagen.helpers import prepare_data_for_leaf
from metagen.metadata import View
from test.fixtures import VIEW, TAG_1, LAYER_TEMPLATE_1
from metagen.metadata import LayerTemplate, Tag
from metagen.components import Timeline, TimelineLegend, TimelineItem, TimelineLayer, State, StateComponents, Maps, \
    ViewComponent, MapDefinitions, Map, MapSetDefinitions, MapSetDefinitionsData, MapSet, MapBackgroundLayer,\
    MapViewSetting, MapSynchronizationSetting
from metagen.presets import PresetBackgroundMaps, PresetMapSynchronization, PresetMapSettingView
from metagen.components import Filter



@pytest.fixture(autouse=True)
def tag():
    return Tag(**prepare_data_for_leaf(TAG_1))


@pytest.fixture(autouse=True)
def layer_template_1():
    return LayerTemplate(**prepare_data_for_leaf(LAYER_TEMPLATE_1))


def test_filter_deserializations():
    assert Filter(**{"layerTemplateKey": "a0c8d936-2e52-4c31-8c55-ffefb1fed8c0"})


def test_set_fillter_from_Leaf(layer_template_1):
    assert Filter.set('layerTemplateKey', layer_template_1)


def test_element_view_model():
    view_obj = prepare_data_for_leaf(VIEW)
    assert View.__wrapped__.parse_obj(view_obj)


def test_set_view_with_timeline_from_sratch(layer_template_1):
    # timeline
    tm1 = Timeline()
    tm1.set_timePeriod(start='2000-01-01', end='2000-02-01')
    legend_1 = TimelineLegend(layerTemplateKey=layer_template_1)
    item_1 = TimelineItem.set_by_layerTemplate(mapZIndex=1, layerTemplate=layer_template_1)
    tm_layer1 = TimelineLayer.set(legend=legend_1, item=item_1)
    tm1.add_timelineLayer(tm_layer1)
    # view
    view_comp = ViewComponent(longDescription='test descrioption')
    #  state components
    state_comp = StateComponents(View=view_comp, Timeline=tm1)
    # maps
    map_def = MapDefinitions(key='map-1')
    map = Map(**{'map-1': map_def})
    map_background = MapBackgroundLayer(**PresetBackgroundMaps.CartoDB_LightNoLabels)
    map_view_setting = MapViewSetting(**PresetMapSettingView.Europe)
    map_sync_setting = MapSynchronizationSetting(**PresetMapSynchronization.AllTrue)
    map_set_data = MapSetDefinitionsData(backgroundLayer=map_background, view=map_view_setting)
    map_set_def = MapSetDefinitions(key='map-set-1', data=map_set_data, maps=['map-1'], sync=map_sync_setting,
                                    activeMapKey= "map-1")
    map_set = MapSet(**{'map-set-1': map_set_def})
    maps =Maps(maps=map, sets=map_set, activeSetKey='map-set')
    # state
    state = State(maps=maps, components=state_comp)
    assert View(nameInternal='test_view',
                applicationKey='test',
                nameDisplay='test',
                description='test description',
                state=state,
                tagKeys=[
            "478e5d14-d564-43d4-b982-978db868b1f0",
            "5ec6b453-046e-4f72-94b3-5f4146fe6224",
            "8fd34a09-bc19-4d97-9d4e-749ac2e8d735",
            "5474a30d-e5f1-4c67-bd12-6174be976893",
            "e08ded4a-9f1c-4aa2-bed7-e5620af10c33"
        ])


def test_set_view_with_timeline_by_methods(tag, layer_template_1):
    # timeline
    time_line = Timeline.set(start='2000-01-01', end='2000-02-01')
    time_line.set_timelineLayer_by_layerTemplate(legend=layer_template_1, items=[layer_template_1])
    # view
    view_comp = ViewComponent(longDescription='test descrioption')
    # maps
    maps = Maps.set(mapKey='map-1', mapSetKey='map-set-1', activeSetKey='map-set', background=PresetBackgroundMaps.CartoDB_LightNoLabels,
                   view_setting=PresetMapSettingView.Europe, map_synchronisation=PresetMapSynchronization.AllTrue)
    # state
    state = State()
    state.add_component(time_line)
    state.add_component(view_comp)
    state.add_maps(maps)

    assert View(nameInternal='test_view',
                applicationKey='test',
                nameDisplay='test',
                description='test description',
                state=state,
                tagKeys=[tag,
            "5ec6b453-046e-4f72-94b3-5f4146fe6224",
            "8fd34a09-bc19-4d97-9d4e-749ac2e8d735",
            "5474a30d-e5f1-4c67-bd12-6174be976893",
            "e08ded4a-9f1c-4aa2-bed7-e5620af10c33"
        ])


import pytest
from metagen.components import Timeline, TimelineLayer, TimelineItem, TimelineLegend
from metagen import LayerTemplate
from test.fixtures import TIME_LAYER_1, TIME_LAYER_2, LAYER_TEMPLATE_1
from datetime import datetime


@pytest.fixture
def time_layer_1():
    return TimelineLayer(**TIME_LAYER_1)


@pytest.fixture
def time_layer_2():
    return TimelineLayer(**TIME_LAYER_2)


@pytest.fixture
def layer_template_1():
    return LayerTemplate(**LAYER_TEMPLATE_1)


# timeline_item
def test_set_timeline_item_by_layer_template(layer_template_1):
    assert TimelineItem.set_by_layerTemplate(mapZIndex=1, layerTemplate=layer_template_1)


# timelineLayer
def test_set_timelinelayer_set(layer_template_1):
    legend = TimelineLegend(layerTemplateKey=layer_template_1)
    item = TimelineItem.set_by_layerTemplate(mapZIndex=1, layerTemplate=layer_template_1)
    assert TimelineLayer.set(legend, item)


# timeline
def test_timeline_set_classmethod():
    tile_layer_1 = TimelineLayer(**TIME_LAYER_1)
    end = datetime(year=2022, month=1, day=1)
    assert Timeline.set(start='2000-01-01', end=end, timelineLayers=tile_layer_1)


def test_timeline_set_classmethod_with_list(time_layer_1, time_layer_2):
    end = datetime(year=2022, month=1, day=1)
    assert Timeline.set(start='2000-01-01', end=end, timelineLayers=[time_layer_1, time_layer_2])


def test_timeline_partial_creation(time_layer_1):
    tm1 = Timeline()
    tm1.set_timePeriod(start='2000-01-01', end='2000-02-01')
    tm1.add_timelineLayer(time_layer_1)
    assert tm1


def test_timeline_partial_creation_with_list(time_layer_1, time_layer_2):
    tm1 = Timeline()
    tm1.set_timePeriod(start='2000-01-01', end='2000-02-01')
    tm1.add_timelineLayer([time_layer_1, time_layer_2])
    assert tm1


def test_timeline_from_scratch(layer_template_1):
    tm1 = Timeline()
    tm1.set_timePeriod(start='2000-01-01', end='2000-02-01')
    legend_1 = TimelineLegend(layerTemplateKey=layer_template_1)
    item_1 = TimelineItem.set_by_layerTemplate(mapZIndex=1, layerTemplate=layer_template_1)
    tm_layer1 = TimelineLayer.set(legend=legend_1,item=item_1)
    tm1.add_timelineLayer(tm_layer1)
    assert tm1



from pydantic import BaseModel, Field
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Union, List, Type
import json

from metagen.base import LeafABC, UUIDEncoder, FactoryABC
from metagen.helpers import create_file, load_json
from metagen.pipes import path_check
from metagen.register import RegisterABC
from metagen.importer import ImporterABC


# serialization & deserialization
class SerializerABC(BaseModel, ABC):

    @abstractmethod
    def to_dict(self):
        pass

    @abstractmethod
    def to_json(self, path: Union[Path, str]) -> None:
        pass


class JSONSerializer(SerializerABC):
    reg: RegisterABC
    structure: dict = Field(default={})

    def to_dict(self) -> dict:
        for element in self.reg.get_elements():
            nodes = element.__nodes__().split('.')
            self.set_node(self.structure, nodes, element)
        return self.structure

    def set_node(self, structure: dict, nodes: list, element: Type[LeafABC]):
        node = nodes.pop(0)
        if len(nodes) > 0:
            if not structure.get(node):
                structure[node] = {}
            self.set_node(structure[node], nodes, element)
        else:
            if not structure.get(node):
                structure[node] = []
            structure[node].append(element.to_dict())

    def to_json(self, path: Union[Path, str]) -> None:
        structure = self.to_dict()

        path = path_check(path)

        if not path.parent:
            create_file(path.parent)

        with open(path, 'w', encoding='utf-8') as file:
            json.dump(structure, file, indent=6, cls=UUIDEncoder, ensure_ascii=False)


class DeSerializerABC(BaseModel, ABC):
    factory: FactoryABC

    @abstractmethod
    def load(self, path: Union[Path, str], **kwargs) -> None:
        pass


class JSONDeserializer(DeSerializerABC):

    def load(self, path: Union[Path, str], encoding='utf8') -> None:

        path = path_check(path)
        obj = load_json(path, encoding)
        for node, structure in obj.items():
            self._parse(node, structure)

    def _parse(self, nodes: str, obj: Union[dict, list]) -> None:

        if isinstance(obj, dict):
            for node, structure in obj.items():
                self._parse(f'{nodes}.{node}', structure)
        elif isinstance(obj, list):
            for data in obj:
                self.factory.create(nodes, data)


# generator
class GeneratorABC(BaseModel, ABC):
    serializer: SerializerABC
    deserializer: DeSerializerABC
    importer: ImporterABC
    reg: RegisterABC

    @abstractmethod
    def import_fixtures(self) -> dict:
        pass

    class Config:
        arbitrary_types_allowed = True


class _Generator(GeneratorABC):
    serializer: SerializerABC
    deserializer: DeSerializerABC
    importer: ImporterABC
    reg: RegisterABC

    def import_fixtures(self) -> dict:
        return self.importer.run(self.to_dict())

    def load_fixtures(self, path: Union[Path, str], encoding='utf8') -> None:
        """Load fixtures into the register"""
        self.deserializer.load(path, encoding=encoding)

    def to_dict(self) -> dict:
        """Generate dict representation of fixtures from register"""
        return self.serializer.to_dict()

    def to_json(self, path: Union[str, Path]) -> None:
        """Generate json representation of fixtures from register"""
        self.serializer.to_json(path)

    def get_element_by_uuid(self, uuid: str) -> Type[LeafABC]:
        """Return element of given uuid"""
        if self.reg.get_by_uuid(uuid):
            return self.reg.get_by_uuid(uuid)
        else:
            raise ValueError(f'Element with uuid {uuid} did not find')

    def get_elements_by_type(self, element: Type[LeafABC]) -> List[Type[LeafABC]]:
        """Return list of all elements of given element type"""
        return [v for k, v in self.reg.uuid.items() if isinstance(v, element.__wrapped__)]



from pydantic import BaseModel, Field, validator
from typing import Optional, Dict, List, Union
from metagen.base import BaseModelWithDynamicKey
from metagen.presets import PresetBackgroundMaps, PresetMapSynchronization, PresetMapSettingView, Preset
from metagen.base import BaseModelArbitrary


class MapDefinitions(BaseModel):
    key: str
    data: dict = Field(default={"view": {}, "layers": []})
    name: Optional[str]


class Map(BaseModelWithDynamicKey):
    __root__: Dict[str, MapDefinitions]


class MapSynchronizationSetting(BaseModelArbitrary):
    roll: Optional[bool]
    tilt: Optional[bool]
    range: Optional[bool]
    center: Optional[bool]
    heading: Optional[bool]
    boxRange: Optional[bool]


class MapBackgroundLayer(BaseModelArbitrary):
    key: str
    type: str
    options: dict


class MapViewSetting(BaseModelArbitrary):
    center: dict
    boxRange: int


class MapSetDefinitionsData(BaseModelArbitrary):
    backgroundLayer: Optional[Union[MapBackgroundLayer, PresetBackgroundMaps]]
    view: Optional[Union[MapViewSetting, PresetMapSettingView]]
    viewLimits: dict = Field(default={})

    @validator('backgroundLayer', pre=True)
    def set_backgroundLayer(cls, value):
        if isinstance(value, dict):
            return MapBackgroundLayer(**value)

    @validator('view', pre=True)
    def set_view(cls, value):
        if isinstance(value, dict):
            return MapViewSetting(**value)


class MapSetDefinitions(BaseModelArbitrary):
    key: str
    data: MapSetDefinitionsData
    maps: List[str]
    sync: Optional[Union[MapSynchronizationSetting, PresetMapSynchronization]]
    activeMapKey: str

    @validator('sync', pre=True)
    def set_sync(cls, value):
        if isinstance(value, dict):
            return MapSynchronizationSetting(**value)
        return value


class MapSet(BaseModelWithDynamicKey):
    __root__: Dict[str, MapSetDefinitions]


class Maps(BaseModel):
    maps: Optional[Map]
    sets: Optional[MapSet]
    activeSetKey: Optional[str]

    @classmethod
    def set(cls, mapKey: str, mapSetKey: str, activeSetKey: str,
            background: Union[MapBackgroundLayer, PresetBackgroundMaps],
            view_setting: Union[MapViewSetting, PresetMapSettingView],
            map_synchronisation: Union[MapSynchronizationSetting, PresetMapSynchronization]):

        maps = Map(**{mapKey: MapDefinitions(key=mapKey)})
        map_set_data = MapSetDefinitionsData(backgroundLayer=background, view=view_setting)
        map_set_def = MapSetDefinitions(key=mapSetKey, data=map_set_data, maps=[mapKey], sync=map_synchronisation,
                                        activeMapKey=mapKey)

        sets = MapSet(**{mapSetKey: map_set_def})
        return cls(maps=maps, sets=sets, activeSetKey=activeSetKey)




from pydantic import BaseModel, Field, root_validator, validator, PrivateAttr
from typing import Optional, Union, List, Literal, Tuple, Type
from pathlib import Path
from inspect import signature
from uuid import UUID, uuid4

from metagen.base import LeafABC, FactoryABC
from metagen.helpers import prepare_data_for_leaf, make_hash, validate_list_uuid
from metagen.components import State
from metagen.main import exist_in_register


# base class
class Leaf(LeafABC):
    key: Optional[Union[UUID, str]] = Field(default_factory=uuid4)
    _input_pars: List[str] = PrivateAttr()

    def __init__(self, **data):
        super().__init__(**data)
        self._input_pars = [k for k in data.keys()]

    def to_dict(self):
        include = {k for k, v in self.__dict__.items() if k in self._input_pars or v is not None}
        exlude = {k for k in self.__dict__.keys()} - include
        data = self.dict(by_alias=True, exclude=exlude, include=include)
        key = data.pop('key')
        return {"key": str(key), "data": data}

    @root_validator(pre=True, allow_reuse=True)
    def set_key(cls, values: dict) -> dict:
        return {k: (v.key if isinstance(v, Leaf) else v) for k, v in values.items()}

    @property
    def hash_attrs(self) -> tuple:
        return tuple()

    def __nodes__(self) -> str:
        pass

    def __hash__(self) -> int:
        return make_hash({k: v for k, v in self.__dict__.items() if k in self.hash_attrs})

    def __setattr__(self, attrName, value):
        super(Leaf, self).__setattr__(attrName, value)

    __slots__ = ('__weakref__',)

    class Config:
        validate_assignment = True


# metadata
@exist_in_register
class Application(Leaf):

    name: str = Field(...)
    nameInternal: Optional[str]
    description: Optional[str]
    color: Optional[str]

    @root_validator()
    def set_application_key(cls, values):
        if values.get('name'):
            values['key'] = values['name']
            return values

    def __nodes__(self) -> str:
        return 'application.applications'

    @property
    def hash_attrs(self) -> tuple:
        return 'name', 'nameInternal'


@exist_in_register
class Configuration(Leaf):
    applicationKey: Optional[Union[str, Leaf]]
    nameInternal: Optional[str]
    data:  Optional[dict]

    def __nodes__(self) -> str:
        return 'application.configurations'

    @property
    def hash_attrs(self) -> tuple:
        return 'applicationKey', 'data', 'nameInternal'


@exist_in_register
class Scope(Leaf):
    applicationKey: Optional[Union[str, Leaf]]
    nameInternal: Optional[str]
    nameDisplay: Optional[str]
    description: Optional[str]
    tagKeys: Optional[Union[List[UUID], List[str]]]
    configuration: Optional[dict]

    _validate_tagKeys = validator('tagKeys', allow_reuse=True)(validate_list_uuid)

    def __nodes__(self) -> str:
        return 'metadata.scopes'

    # FIXME: add hashable configuration
    @property
    def hash_attrs(self) -> tuple:
        return 'applicationKey', 'nameDisplay', 'tagsKey', 'configuration', 'nameInternal'


@exist_in_register
class Place(Leaf):
    applicationKey: Optional[Union[str, Leaf]]
    scopeKey: Optional[UUID]
    nameInternal: Optional[str]
    nameDisplay: Optional[str]
    description: Optional[str]
    geometry: Optional[Union[dict, Path]]
    bbox: Optional[List[float]]

    def __nodes__(self) -> str:
        return 'metadata.places'

    @property
    def hash_attrs(self) -> tuple:
        return 'applicationKey', 'nameDisplay', 'scopeKey', 'nameInternal'


@exist_in_register
class LayerTemplate(Leaf):
    applicationKey: Optional[Union[str, Leaf]]
    nameInternal: Optional[str]
    nameDisplay: Optional[str]
    description: Optional[str]
    scopeKey: Optional[Union[UUID, Leaf]]
    tagKeys: Optional[Union[List[UUID], List[str]]]

    _validate_tagKeys = validator('tagKeys', allow_reuse=True)(validate_list_uuid)

    def __nodes__(self) -> str:
        return 'metadata.layerTemplates'

    @property
    def hash_attrs(self) -> tuple:
        return 'applicationKey', 'nameDisplay', 'nameInternal'


@exist_in_register
class Attribute(Leaf):
    applicationKey: Optional[Union[str, Leaf]]
    nameInternal: Optional[str]
    nameDisplay: Optional[str]
    description: Optional[str]
    type: Optional[Literal['number', 'text', 'bool']]
    unit: Optional[str]
    valueType: Optional[str]
    color: Optional[str]

    def __nodes__(self) -> str:
        return 'metadata.attributes'

    @property
    def hash_attrs(self) -> tuple:
        return 'applicationKey', 'nameDisplay', 'type', 'nameInternal', 'unit', 'nameInternal'


@exist_in_register
class Period(Leaf):
    applicationKey: Optional[Union[str, Leaf]]
    scopeKey: Optional[UUID]
    nameInternal: Optional[str]
    nameDisplay: Optional[str]
    description: Optional[str]
    start: Optional[str]
    end: Optional[str]
    period: Optional[str]
    tagKeys: Optional[Union[List[UUID], List[str]]]

    _validate_tagKeys = validator('tagKeys', allow_reuse=True)(validate_list_uuid)

    @root_validator(pre=True)
    def handle_prtData_input(cls, values):
        if all([key not in values for key in ['period', 'start', 'end']]):
            raise ValueError('Period has to have period or start and end parameters')
        elif 'period' not in values and any([key not in values for key in ['start', 'end']]):
            raise ValueError('Period has to have defined start and end parameters')
        else:
            return values

    def __nodes__(self) -> str:
        return 'metadata.periods'

    @property
    def hash_attrs(self) -> tuple:
        return 'applicationKey', 'nameDisplay', 'period', 'start', 'end', 'nameInternal', 'tagKeys'


@exist_in_register
class Case(Leaf):
    applicationKey: Optional[Union[str, Leaf]]
    nameInternal: Optional[str]
    nameDisplay: Optional[str]
    description: Optional[str]
    tagKeys: Optional[Union[List[UUID], List[str]]]

    _validate_tagKeys = validator('tagKeys', allow_reuse=True)(validate_list_uuid)

    def __nodes__(self) -> str:
        return 'metadata.cases'

    @property
    def hash_attrs(self) -> tuple:
        return 'applicationKey', 'nameDisplay', 'nameInternal', 'tagKeys'


@exist_in_register
class Style(Leaf):
    nameDisplay: Optional[str]
    nameInternal: Optional[str]
    description: Optional[str]
    source: Optional[str]
    nameGeoserver: Optional[str]
    definition: Optional[dict]
    applicationKey: Optional[Union[str, Leaf]]
    tagKeys: Optional[UUID]

    def __nodes__(self) -> str:
        return 'metadata.styles'

    @property
    def hash_attrs(self) -> tuple:
        return 'nameDisplay', 'description', 'source', 'nameGeoserver', 'applicationKey', 'tagKeys', 'definition', \
               'nameInternal'


@exist_in_register
class Tag(Leaf):
    nameInternal: Optional[str]
    nameDisplay: Optional[str]
    description: Optional[str]
    color: Optional[str]
    tagKeys: Optional[Union[List[UUID], List[str], List[Type[LeafABC]]]]

    _validate_tagKeys = validator('tagKeys', allow_reuse=True)(validate_list_uuid)

    def __nodes__(self) -> str:
        return 'metadata.tags'

    @property
    def hash_attrs(self) -> tuple:
        return 'nameDisplay', 'tagKeys', 'nameInternal'


# datasource


@exist_in_register
class SpatialVector(Leaf):
    nameInternal: Optional[str]
    layerName: Optional[str]
    tableName: Optional[str]
    fidColumnName: Optional[str] = Field(default='ogc_fid')
    geometryColumnName: Optional[str] = Field(default='geom')
    type: Literal['vector'] = Field(default='vector')

    def __nodes__(self) -> str:
        return 'dataSources.spatial'

    @property
    def hash_attrs(self) -> tuple:
        return 'type', 'layerName', 'tableName', 'fidColumnName', 'geometryColumnName', 'nameInternal'


@exist_in_register
class SpatialWMS(Leaf):
    nameInternal: Optional[str]
    url: Optional[str]
    layers: Optional[str]
    styles: Optional[str]
    configuration: Optional[dict]
    params: Optional[dict]
    attribution: Optional[str]
    type: Literal['wms'] = Field(default='wms')

    def __nodes__(self) -> str:
        return 'dataSources.spatial'

    @property
    def hash_attrs(self) -> tuple:
        return 'url', 'layers', 'styles', 'params', 'configuration', 'type', 'attribution', 'nameInternal'


@exist_in_register
class SpatialWMTS(Leaf):
    nameInternal: Optional[str]
    urls: Optional[List[str]]
    type: Literal['wmts'] = Field(default='wmts')

    def __nodes__(self) -> str:
        return 'dataSources.spatial'

    @property
    def hash_attrs(self) -> tuple:
        return 'urls', 'type'


@exist_in_register
class SpatialCOG(Leaf):
    nameInternal: Optional[str]
    url: Optional[str]
    type: Literal['cog'] = Field(default='cog')

    def __nodes__(self) -> str:
        return 'dataSources.spatial'

    @property
    def hash_attrs(self) -> tuple:
        return 'url', 'type', 'nameInternal'


@exist_in_register
class SpatialAttribute(Leaf):
    """
    Spatial attribute is used to create link on data stored in PostgreSQL Database.
    Element attribute:
     - nameInternal: Optional[str] : internal name of element, unique
     - attribution: Optional[str] : '''todo: add description'''
     - tableName: Optional[str] : PostgreSQL table name
     - columnName: Optional[str] : PostgreSQL column name
     - fidColumnName: Optional[str] : PostgreSQL fid column name
    """
    nameInternal: Optional[str]
    attribution: Optional[str]
    tableName: Optional[str]
    columnName: Optional[str]
    fidColumnName: Optional[str] = Field(default='ogc_fid')

    def __nodes__(self) -> str:
        return 'dataSources.attribute'

    @property
    def hash_attrs(self) -> tuple:
        return 'attribution', 'tableName', 'columnName', 'fidColumnName', 'nameInternal'


# relations
@exist_in_register
class RelationSpatial(Leaf):
    nameInternal: Optional[str]
    scopeKey: Optional[Union[UUID, Leaf]]
    periodKey: Optional[Union[UUID, Leaf]]
    placeKey: Optional[Union[UUID, Leaf]]
    spatialDataSourceKey: Optional[Union[UUID, Leaf]]
    layerTemplateKey: Optional[Union[UUID, Leaf]]
    applicationKey: Optional[Union[str, Leaf]]
    caseKey: Optional[Union[UUID, Leaf]]
    scenarioKey: Optional[Union[UUID, Leaf]]

    def __nodes__(self) -> str:
        return 'relations.spatial'

    @property
    def hash_attrs(self) -> tuple:
        return 'scopeKey:', 'periodKey', 'placeKey', 'spatialDataSourceKey', 'layerTemplateKey', 'applicationKey', \
               'caseKey', 'scenarioKey', 'nameInternal'


@exist_in_register
class RelationAttribute(Leaf):
    nameInternal: Optional[str]
    scopeKey: Optional[Union[UUID, Leaf]]
    periodKey: Optional[Union[UUID, Leaf]]
    placeKey: Optional[Union[UUID, Leaf]]
    attributeDataSourceKey: Optional[Union[UUID, Leaf]]
    layerTemplateKey: Optional[Union[UUID, Leaf]]
    scenarioKey: Optional[Union[UUID, Leaf]]
    caseKey: Optional[Union[UUID, Leaf]]
    attributeSetKey: Optional[Union[UUID, Leaf]]
    attributeKey: Optional[Union[UUID, Leaf]]
    areaTreeLevelKey: Optional[Union[UUID, Leaf]]
    applicationKey: Optional[Union[str, Leaf]]

    def __nodes__(self) -> str:
        return 'relations.attribute'

    @property
    def hash_attrs(self) -> tuple:
        return 'scopeKey:', 'periodKey', 'placeKey', 'attributeDataSourceKey', 'layerTemplateKey', 'scenarioKey', \
               'caseKey', 'attributeSetKey', 'attributeKey', 'areaTreeLevelKey', 'applicationKey', 'nameInternal'

# views


@exist_in_register
class View(Leaf):
    nameInternal: Optional[str]
    applicationKey: Optional[Union[str, Leaf]]
    nameDisplay: Optional[str]
    description: Optional[str]
    state: Optional[State]
    tagKeys: Optional[List[Union[UUID, str, Leaf]]]

    def __nodes__(self) -> str:
        return 'views.views'

    @property
    def hash_attrs(self) -> tuple:
        return 'applicationKey', 'nameDisplay', 'state', 'tagKeys', 'nameInternal'

    @validator('tagKeys', pre=True)
    def set_tagKey(cls, values):
        return [value.key if isinstance(value, Leaf) else value for value in values]


class ElementSignature(BaseModel):
    parameters: Tuple[str, ...]
    type: Optional[str]

    @root_validator(pre=True)
    def get_pars_from_signature(cls, values)-> tuple:
        element = values.get('parameters')
        try:
            values['parameters'] = tuple([pars for pars in signature(element).parameters])
        except TypeError:
            values['parameters'] = tuple([pars for pars in signature(element.__wrapped__).parameters])

        try:
            values['type'] = element.__fields__.get('type').default
        except AttributeError:
            pass

        return values

    def check_signature(self, input_parameters: dict) -> bool:
        if self.type:
            if self.type != input_parameters.get('type'):
                return False

        return all([True if k in self.parameters else False for k in input_parameters.keys()])

    def __hash__(self):
        return hash(self.parameters)


class ElementFactory(FactoryABC, BaseModel):
    elements_register: dict = Field(default_factory=dict)

    def add(self, element: Type[LeafABC]) -> None:
        element_signature = ElementSignature(parameters=element)

        if not self.elements_register.get(element.__nodes__(self)):
            self.elements_register.update({element.__nodes__(self): {}})

        self.elements_register[element.__nodes__(self)].update({element_signature: element})

    def _get_element_types(self, nodes: str) -> dict:
        try:
            return self.elements_register[nodes]
        except KeyError:
            raise NotImplementedError(f'No elements for node {nodes}')

    @staticmethod
    def _find_element_by_signature(element_types: dict, init_data: dict)-> Optional[Type[LeafABC]]:
        for signature, element in element_types.items():
            if signature.check_signature(init_data):
                return element
        return None

    def create(self, nodes: str, data: dict) -> Type[LeafABC]:
        element_types = self._get_element_types(nodes)
        init_data = prepare_data_for_leaf(data)
        element = self._find_element_by_signature(element_types, init_data)
        if element:
            return element(**init_data)
        else:
            raise ValueError(f'No element signature match input element parameters '
                             f'{", ".join([par for par in init_data])} and nodes: {nodes} ')








element_factory = ElementFactory()
element_factory.add(Application)
element_factory.add(Configuration)
element_factory.add(Scope)
element_factory.add(Place)
element_factory.add(LayerTemplate)
element_factory.add(Attribute)
element_factory.add(Period)
element_factory.add(Case)
element_factory.add(Style)
element_factory.add(SpatialVector)
element_factory.add(SpatialWMS)
element_factory.add(SpatialWMTS)
element_factory.add(SpatialAttribute)
element_factory.add(RelationSpatial)
element_factory.add(RelationAttribute)
element_factory.add(View)
element_factory.add(Tag)
element_factory.add(SpatialCOG)



# general and helper function and classe

from typing import Union, Optional, Any, Type
from pydantic import BaseModel
from pydantic.utils import ROOT_KEY
from abc import ABC, abstractmethod
from uuid import UUID
import json


# helper class
class SingletonMeta(type):
    _instance = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instance:
            cls._instance[cls] = super(SingletonMeta, cls).__call__(*args, **kwargs)
        return cls._instance[cls]


class BaseModelWithDynamicKey(BaseModel):
    """
    Pydantic workaoround for custom dynamic key
    ref: https://stackoverflow.com/questions/60089947/creating-pydantic-model-schema-with-dynamic-key
    """

    def __init__(self, **data: Any) -> None:
        if self.__custom_root_type__ and data.keys() != {ROOT_KEY}:
            data = {ROOT_KEY: data}
        super().__init__(**data)


class BaseModelArbitrary(BaseModel):
    pass

    class Config:
        arbitrary_types_allowed = True


# model classes
class LeafABC(BaseModel, ABC):
    key: Optional[UUID]

    @abstractmethod
    def __nodes__(self) -> str:
        pass

    @property
    @abstractmethod
    def hash_attrs(self) -> tuple:
        pass


class FactoryABC(BaseModel, ABC):

    @classmethod
    def add(self, element: Any) -> None:
        pass

    @classmethod
    def create(self, nodes: str, data: dict) -> Any:
        pass


def set_key_from_input(value: Union[str, UUID, Type[LeafABC]]):
    """
    Helper method used as validator in pydantic model.
    For input string, Leaf or UUID return valid UUID
    """
    if isinstance(value, LeafABC):
        return value.key
    if isinstance(value, str):
        return UUID(value)
    return value


# helper class
class UUIDEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, UUID):
            # if the obj is uuid, we simply return the value of uuid
            return str(obj)
        return json.JSONEncoder.default(self, obj)




import pytest

from metagen.metadata import Application, LayerTemplate
from .fixtures.elements import layer_template_1a, layer_template_1b, application_1a, configuration_1a, scope_1a, \
    scope_1b, place_1a, period_1a, attribute_1a, spatial_wms_1a, case_1a, relation_spatial_1a, spatial_attribute_1a, \
    spatial_wmts_1a, spatial_vector_1a, style_1a, tag_1a, relation_attribute_1a


def test_element_duplicity(layer_template_1a, layer_template_1b):
    assert layer_template_1a.key == layer_template_1b.key


def test_hashbility(layer_template_1a, application_1a, configuration_1a, scope_1a, place_1a, period_1a, attribute_1a,
                    spatial_wms_1a, case_1a, relation_spatial_1a, spatial_attribute_1a, spatial_wmts_1a,
                    spatial_vector_1a, style_1a, tag_1a, relation_attribute_1a):
    assert hash(layer_template_1a)
    assert hash(application_1a)
    assert hash(configuration_1a)
    assert hash(scope_1a)
    assert hash(place_1a)
    assert hash(period_1a)
    assert hash(attribute_1a)
    assert hash(spatial_wms_1a)
    assert hash(case_1a)
    assert hash(relation_spatial_1a)
    assert hash(spatial_attribute_1a)
    assert hash(spatial_wmts_1a)
    assert hash(spatial_vector_1a)
    assert hash(style_1a)
    assert hash(tag_1a)
    assert hash(relation_attribute_1a)


def test_hash_independence_on_order_of_tagKeys(scope_1a,scope_1b):
    assert hash(scope_1a) == hash(scope_1b)


def test_stac_attribute_in_leaf():
    app = Application(name='test', nameInternal='test')
    assert hasattr(app, '_input_pars')
    assert app._input_pars == ['name', 'nameInternal']


def test_exlude_stac_attribute_from_leaf():
    app = Application(name='test', nameInternal='test')
    dict = app.dict()
    assert dict.get('_input_pars') is None


def test_to_dict():
    """testing of exluding atributes with none values except those specified as none during the inicialization"""
    lt = LayerTemplate(applicationKey=None, nameInternal='test')
    lt_dict = lt.to_dict()
    data = lt_dict.get('data')
    assert all([k in ['applicationKey', 'nameInternal'] for k in data.keys()])
    assert data.get('applicationKey') is None


# def test_refresh_hash():
#     lt1 = LayerTemplate(applicationKey='app_test', nameInternal=f'lt', nameDisplay='lt')
#     lt2 = LayerTemplate(applicationKey='app', nameInternal=f'lt', nameDisplay='lt')
#     lt2.nameDisplay= 'app_test'
#     assert hash(lt1) == hash(lt2)



from pydantic import BaseModel, Field
from typing import Optional


from ._map import Maps, MapViewSetting, MapSynchronizationSetting, MapBackgroundLayer, MapDefinitions, Map, \
    MapSetDefinitions, MapSet, MapSetDefinitionsData
from ._timeline import Timeline, TimelineLayer, TimelineItem, TimelineLegend
from ._view import ViewComponent
from ._general import Filter, Component


class StateComponents(Component):
    View: Optional[ViewComponent]
    Timeline: Optional[Timeline]

    def add_component(self, component: Component):
        for k, v in self.__annotations__.items():
            if isinstance(component, v.__args__[0]):
                setattr(self, k, component)


class State(Component):
    maps: Optional[Maps]
    components: Optional[StateComponents] = Field(default_factory=StateComponents)

    def add_maps(self, maps: Maps):
        self.maps = maps

    def add_component(self, component: Component):
        self.components.add_component(component)







from pathlib import Path
from typing import Optional, Union, List, Literal, Callable
from uuid import UUID, uuid4

from pydantic import Field, PrivateAttr, BaseModel, root_validator

from metagen.components import State
import pydantic.main


class Leaf(BaseModel):
    key: Optional[Union[UUID, str]] = Field(default_factory=uuid4)

    class Config:
        validate_assignment = True
        allow_mutation = True

    def __setattr__(self, key, value):
        return object.__setattr__(self, key, value)

    def dict(self, **ignore):
        attr_dict = self.__dict__.copy()
        key = attr_dict.pop('key')
        return {'key': str(key), 'data': {k: v for k,v in attr_dict.items() if v is not None}}

    @root_validator(pre=True)
    def prepare_date(cls, values: dict):
        if values is not None:
            data = values.pop('data')
            values.update({k: v for k, v in data.items()})
            return values
        else:
            return values


class Node(BaseModel):

    @root_validator(pre=True)
    def prepare_date(cls, values: dict):
        if values is not None:
            return values


class Application(Leaf):
    name: Optional[str]
    nameInternal: Optional[str]
    description: Optional[str]
    color: Optional[str]


class Configuration(Leaf):
    applicationKey: Optional[Union[str, Leaf]]
    nameInternal: Optional[str]
    data:  Optional[dict]


class Applications(Node):
    applications: Optional[List[Application]]
    configurations: Optional[List[Configuration]]


class Scope(Leaf):
    applicationKey: Optional[Union[str, Leaf]]
    nameInternal: Optional[str]
    nameDisplay: Optional[str]
    description: Optional[str]


class Place(Leaf):
    applicationKey: Optional[Union[str, Leaf]]
    scopeKey: Optional[UUID]
    nameInternal: Optional[str]
    nameDisplay: Optional[str]
    description: Optional[str]
    geometry: Optional[Union[dict, Path]]
    bbox: Optional[List[float]]


class LayerTemplate(Leaf):
    applicationKey: Optional[Union[str, Leaf]]
    nameInternal: Optional[str]
    nameDisplay: Optional[str]
    description: Optional[str]


class Attribute(Leaf):
    applicationKey: Optional[Union[str, Leaf]]
    nameInternal: Optional[str]
    nameDisplay: Optional[str]
    description: Optional[str]
    type: Optional[Literal['numeric', 'text', 'bool']]
    unit: Optional[str]
    valueType: Optional[str]
    color: Optional[str]


class Period(Leaf):
    applicationKey: Optional[Union[str, Leaf]]
    scopeKey: Optional[UUID]
    nameInternal: Optional[str]
    nameDisplay: Optional[str]
    description: Optional[str]
    start: Optional[str]
    end: Optional[str]
    period: Optional[str]


class Case(Leaf):
    applicationKey: Optional[Union[str, Leaf]]
    nameInternal: Optional[str]
    nameDisplay: Optional[str]
    description: Optional[str]


class Style(Leaf):
    nameDisplay: Optional[str]
    nameInternal: Optional[str]
    description: Optional[str]
    source: Optional[str]
    nameGeoserver: Optional[str]
    definition: Optional[dict]
    applicationKey: Optional[Union[str, Leaf]]
    tagKeys: Optional[UUID]


class Tag(Leaf):
    nameInternal: Optional[str]
    nameDisplay: Optional[str]
    description: Optional[str]
    color: Optional[str]
    tagKeys: Optional[List[UUID]]


class Metadata(Node):
    scopes: Optional[List[Scope]]
    places: Optional[List[Place]]
    layerTemplates: Optional[List[LayerTemplate]]
    attributes: Optional[List[Attribute]]
    periods: Optional[List[Period]]
    styles: Optional[List[Style]]
    tags: Optional[List[Tag]]
    cases: Optional[List[Case]]


# datasource
class SpatialVector(Leaf):
    nameInternal: Optional[str]
    layerName: Optional[str]
    tableName: Optional[str]
    fidColumnName: Optional[str] = Field(default='ogc_fid')
    geometryColumnName: Optional[str] = Field(default='geom')
    type: Literal['vector'] = Field(default='vector')


class SpatialWMS(Leaf):
    nameInternal: Optional[str]
    url: Optional[str]
    layers: Optional[str]
    styles: Optional[str]
    configuration: Optional[dict]
    params: Optional[dict]
    type: Literal['wms'] = Field(default='wms')


class SpatialWMTS(Leaf):
    nameInternal: Optional[str]
    urls: Optional[List[str]]
    type: Optional[Literal['wmts']] = Field(default='wmts')


class SpatialCOG(Leaf):
    nameInternal: Optional[str]
    url: Optional[str]
    type: Optional[Literal['cog']] = Field(default='cog')


class SpatialAttribute(Leaf):
    nameInternal: Optional[str]
    attribution: Optional[str]
    tableName: Optional[str]
    columnName: Optional[str]
    fidColumnName: Optional[str] = Field(default='ogc_fid')


class DataSources(Node):
    spatial: Optional[List[Union[SpatialWMS,SpatialWMTS,SpatialCOG,SpatialVector]]]
    attribute: Optional[List[SpatialAttribute]]


class RelationSpatial(Leaf):
    nameInternal: Optional[str]
    scopeKey: Optional[Union[UUID, Leaf]]
    periodKey: Optional[Union[UUID, Leaf]]
    placeKey: Optional[Union[UUID, Leaf]]
    spatialDataSourceKey: Optional[Union[UUID, Leaf]]
    layerTemplateKey: Optional[Union[UUID, Leaf]]
    applicationKey: Optional[Union[str, Leaf]]
    caseKey: Optional[Union[UUID, Leaf]]


class RelationAttribute(Leaf):
    nameInternal: Optional[str]
    scopeKey: Optional[Union[UUID, Leaf]]
    periodKey: Optional[Union[UUID, Leaf]]
    placeKey: Optional[Union[UUID, Leaf]]
    attributeDataSourceKey: Optional[Union[UUID, Leaf]]
    layerTemplateKey: Optional[Union[UUID, Leaf]]
    scenarioKey: Optional[Union[UUID, Leaf]]
    caseKey: Optional[Union[UUID, Leaf]]
    attributeSetKey: Optional[Union[UUID, Leaf]]
    attributeKey: Optional[Union[UUID, Leaf]]
    areaTreeLevelKey: Optional[Union[UUID, Leaf]]
    applicationKey: Optional[Union[str, Leaf]]


class Relations(Node):
    spatial: Optional[List[RelationSpatial]]
    attribute: Optional[List[RelationAttribute]]


class View(Leaf):
    nameInternal: Optional[str]
    applicationKey: Optional[Union[str, Leaf]]
    nameDisplay: Optional[str]
    description: Optional[str]
    state: Optional[dict]
    tagKeys: Optional[List[Union[UUID, str, Leaf]]]



class Views(Node):
    views: Optional[List[View]]


class Structure(Node):
    application: Optional[Applications]
    metadata: Optional[Metadata]
    dataSources: Optional[DataSources]
    relations: Optional[Relations]
    views: Optional[Views]


if __name__ == '__main__':
    from pathlib import Path
    from metagen.helpers import load_json
    import re
    file = Path('C:\\Users\micha\PycharmProjects\\app-esaWorldWater\\fixtures.json')
    fixtures = load_json(file)
    fix = Structure(**fixtures)

    def get_keys(array: List[Leaf], keys: dict):
        for item in array:
            keys.update({item.key: item})

    def get_relationKeys(array: List[Leaf], keys: dict):
        for item in array:
            for k, v in item.__dict__.items():
                if k.__contains__('Key'):
                    keys.update({v: item})

    def loop_structure2get(structure: Structure, keys: dict, func: Callable):
        for k, v in structure.__dict__.items():
            if v is not None:
                if isinstance(v, list):
                    func(v, keys)
                else:
                    loop_structure2get(v, keys, func)
        return keys

    def find_in(structure: list, attr, match):
        for s in structure:
            if getattr(s, attr) == match:
                return s


    key_list = loop_structure2get(fix, keys={}, func=get_keys)
    relkey_list = loop_structure2get(fix, keys={}, func=get_relationKeys)
    missed = set([v for v in relkey_list]).difference(set([v for v in key_list]))
    for mis in missed:
        print(mis)
    print(f'total: {len(missed)}')

    print('oprava:')
    for mis in missed:
        print(mis)
        element = relkey_list[mis]
        if element.nameInternal.__contains__('attributerelation'):
            matches = re.findall('([\d]{4}_[\d]{2}_[\d]{2})', element.nameInternal)

            period_str = '-'.join(matches[0].split('_'))
            periodElement =find_in(fix.metadata.periods, 'period', period_str)
            element.periodKey = periodElement.key

    new_fix = fix.dict(exclude_none=True)

    import json
    from metagen.base import UUIDEncoder

    with open(file.parent / 'new_fixtures.json', 'w') as jfile:
        json.dump(new_fix, jfile, indent=6, cls=UUIDEncoder)






from pydantic import BaseModel, validator, AnyUrl
from pathlib import Path
import subprocess
from requests import post
from abc import ABC, abstractmethod

from metagen.pipes import path_check


class ImportError(Exception):
    pass


class ImporterABC(BaseModel, ABC):

    @abstractmethod
    def run(self, fixtutres: dict) -> dict:
        pass


class Importer(ImporterABC):
    path: Path
    instance_url: AnyUrl
    host: AnyUrl

    @validator('path', pre=True)
    def set_pat(cls, value):
        return path_check(value)

    def run(self, fixtutres: dict) -> dict:
        self.server_start()
        report = self.import_fixtures(fixtutres)
        self.server_stop()
        return report

    def server_start(self):
        """Start importer node server"""
        subprocess.run(f'node {self.path / "server"} start')

    def server_stop(self):
        """Start importer node server"""
        subprocess.run(f'node {self.path / "server"} stop')

    def import_fixtures(self, fixtures: dict):
        response = post(url=f'{self.host}/metadata', data=fixtures)
        if response.status_code == 200:
            return response.json()
        else:
            raise ImportError(f'Import of fixtures failed '
                              f'\n status: {response.status_code} '
                              f'\n message: {response.content}')

